import os
from contextlib import contextmanager

from ocs_archive.storage.filestore import FileStore
from ocs_archive.input.file import DataFile
from ocs_archive.settings import settings

class FileSystemStore(FileStore):
    """This class stores the files locally on the file system, in the base directory specified.

    It does not support versioning of files, and will overwrite any file
    with the same path and filename. Please use S3 storage if you want versioning support.
    """
    def __init__(self, root_dir: str = settings.FILESYSTEM_STORAGE_ROOT_DIR):
        """Create filesystem storage manager using the root directory specified."""
        super().__init__()
        self.root_dir = root_dir

    def store_file(self, data_file: DataFile):
        md5 = data_file.open_file.get_md5()
        directory = os.path.join(self.root_dir, os.path.dirname(data_file.get_filestore_path()))
        os.makedirs(directory, exist_ok=True)  # ensure that the directory exists or make it
        with open(os.path.join(self.root_dir, data_file.get_filestore_path()), 'wb') as fp:
            fp.write(data_file.open_file.get_from_start().read())
        return {
            'key': md5,
            'md5': md5,
            'extension': data_file.open_file.extension
        }

    def delete_file(self, path: str, version_id: str):
        full_path = os.path.join(self.root_dir, path)
        if os.path.exists(full_path):
            os.remove(full_path)

    def get_url(self, path: str, version_id: str, expiration: float):
        # Filesystem storage files are assumed to be hosted at their base directory on the base url
        return os.path.join(settings.FILESYSTEM_STORAGE_BASE_URL, path)

    @contextmanager
    def get_fileobj(self, path: str):
        full_path = os.path.join(self.root_dir, path)
        fileobj = None
        try:
            if os.path.exists(full_path):
                fileobj = open(full_path, 'rb')
            yield fileobj
        finally:
            if fileobj:
                fileobj.close()

    def get_file_size(self, path: str):
        full_path = os.path.join(self.root_dir, path)
        return os.path.getsize(full_path)



from ocs_archive.input.file import FileSpecificationException
from ocs_archive.input.tarwithfitsfile import TarWithFitsFile
from ocs_archive.input.lcofitsfile import LcoFitsFile


class LcoTarWithFitsFile(TarWithFitsFile, LcoFitsFile):
    @staticmethod
    def _get_meta_file_from_targz(tarfileobj):
        for member in tarfileobj.getmembers():
            if any(x + '.fits' in member.name for x in ['e00', 'e90', 'e91']) and member.isfile():
                return tarfileobj.extractfile(member)
        raise FileSpecificationException('Tar file missing meta fits!')



from datetime import timedelta
import hashlib
import os
import io
from urllib.parse import urljoin
import itertools

from dateutil.parser import parse
from astropy import wcs, units
from astropy.coordinates import Angle
import requests
from functools import lru_cache
from requests.models import HTTPError

from ocs_archive.settings import settings
from ocs_archive.input.headerdata import HeaderData


class FileSpecificationException(Exception):
    pass


class File:
    """Operates on an open file-like object to extract its extension and retreive its metadata."""
    def __init__(self, fileobj, path=None):
        """Create a File from a file like object and optional string path."""
        self.fileobj = fileobj
        self.path = path
        self.basename, self.extension = self.get_basename_and_extension(self.filename)

    def get_from_start(self):
        self.fileobj.seek(0)
        return self.fileobj

    def get_md5(self):
        return hashlib.md5(self.get_from_start().read()).hexdigest()

    def __len__(self):
        """Returns the length of the file like object in bytes."""
        self.fileobj.seek(0, os.SEEK_END)
        length = self.fileobj.tell()
        self.fileobj.seek(0)
        return length

    @property
    def filename(self):
        # The filename, can be a full path or at least the basename plus the extension
        filename = None
        if self.path is not None:
            filename = self.path
        elif hasattr(self.fileobj, 'name'):
            filename = self.fileobj.name
        elif hasattr(self.fileobj, 'filename'):
            filename = self.fileobj.filename
        return filename

    @staticmethod
    def get_basename_and_extension(path):
        basename, extension = None, None
        if path is not None:
            filename = os.path.basename(path)
            if filename.find('.') > 0:
                basename = filename[:filename.index('.')]
                extension = filename[filename.index('.'):]
            else:
                basename = filename
                extension = ''
        return basename, extension


class EmptyFile(File):
    """Implements the file methods but has no open file, just the supplied filename."""
    def __init__(self, path):
        """Create an empty 0 byte File with the given path name."""
        super().__init__(io.BytesIO(b''), path=path)

    def get_md5(self):
        return hashlib.md5(self.filename).hexdigest()

    @property
    def filename(self):
        # The filename, can be a full path or at least the basename plus the extension
        filename = 'dummy_file.fits'
        if self.path is not None:
            filename = self.path
        return filename


class DataFile:
    """
    Base class for extracting metadata needed to store and filter data from an input file-like object.
    This class should be subclassed and methods overriden to support other input data formats in the archive.
    """
    def __init__(self, open_file: File, file_metadata: dict = None, blacklist_headers: tuple = settings.HEADER_BLACKLIST, required_headers: tuple = settings.REQUIRED_HEADERS):
        """Create a DataFile with an open File object and optional metadata."""
        self.open_file = open_file
        self.blacklist_headers = blacklist_headers
        self.required_headers = required_headers
        if file_metadata is None:
            file_metadata = {}
        self._create_header_data(file_metadata)
        self._repair_observation_day()
        self._repair_public_date()

    @property
    @lru_cache(maxsize=1000)
    def proposal_tags(self):
        """Return the tags associated with a proposal in the Observation Portal."""
        # If the observation portal connection isn't configured, don't try and fetch proposal tags
        if None in [settings.OBSERVATION_PORTAL_BASE_URL, settings.OBSERVATION_PORTAL_API_TOKEN]:
            return None

        proposal_id = self.header_data.get_proposal_id()
        proposal_url = urljoin(settings.OBSERVATION_PORTAL_BASE_URL, f'/api/proposals/{proposal_id}')
        try:
            response = requests.get(proposal_url,
                                    headers={'Authorization': f'Token {settings.OBSERVATION_PORTAL_API_TOKEN}'})
            response.raise_for_status()
        except HTTPError:
            return None
        return response.json()['tags']

    @property
    def data_privacy_tags(self):
        """Given a set of proposal tags, return tags that match tags defined as data privacy tags

        These are tags defined in settings.PRIVATE_PROPOSAL_TAGS and settings.PUBLIC_PROPOSAL_TAGS
        """
        if self.proposal_tags is not None:
            privacy_tags = [tag for tag in self.proposal_tags if tag in itertools.chain(settings.PRIVATE_PROPOSAL_TAGS, settings.PUBLIC_PROPOSAL_TAGS)]
            return privacy_tags
        else:
            return []

    def _create_header_data(self, file_metadata: dict):
        if self._is_valid_file_metadata(file_metadata):
            self.header_data = HeaderData(file_metadata)
            return
        # Missing one or more required headers in the input file_metadata
        raise FileSpecificationException('Could not find required keywords in headers!')

    def get_header_data(self):
        # This should return the header values you want to store in the archive.
        # For a fits file, this would be the dictionary of all header values
        return self.header_data

    def _repair_observation_day(self):
        """
        Set observation day from the headers of date
        :return: Day in YYYYMMDD format
        """
        if not self.header_data.get_observation_day():
            observation_date = self.header_data.get_observation_date()
            if observation_date:
                observation_day = observation_date.split('T')[0].replace('-', '')
                self.header_data.update_headers({settings.OBSERVATION_DAY_KEY: observation_day})

    def _repair_public_date(self):
        # Set the public date based on observation date. Should be overriden if you have another method
        # of specifying the public date in your file type
        if not self.header_data.get_public_date():
            # Check if the proposal is denoted as public or private in the observation portal
            if any([tag in settings.PRIVATE_PROPOSAL_TAGS for tag in self.data_privacy_tags]):
                public_date = (parse(self.header_data.get_observation_date()) + timedelta(days=365 * 999)).isoformat()
            elif any([tag in settings.PUBLIC_PROPOSAL_TAGS for tag in self.data_privacy_tags]):
                public_date = self.header_data.get_observation_date()
            else:
                # If it's a calibration frame, it's immediately pubic
                if (self.header_data.get_configuration_type() in settings.CALIBRATION_TYPES or
                        (self.header_data.get_proposal_id() and any([prop in self.header_data.get_proposal_id() for prop in settings.PUBLIC_PROPOSALS]))):
                    public_date = self.header_data.get_observation_date()
                # If the proposal is set to private or the file type is a private file type, make it private forever
                elif ((self.header_data.get_proposal_id() and any([prop in self.header_data.get_proposal_id() for prop in settings.PRIVATE_PROPOSALS])) or
                        any([chars in self.open_file.basename for chars in settings.PRIVATE_FILE_TYPES])):
                    public_date = (parse(self.header_data.get_observation_date()) + timedelta(days=365 * 999)).isoformat()
                # Finally, if none of these, make it proprietary
                else:
                    # This should be proprietary, set it to X days from observation date
                    public_date = (parse(self.header_data.get_observation_date()) + timedelta(days=settings.DAYS_UNTIL_PUBLIC)).isoformat()
            
            self.header_data.update_headers({settings.PUBLIC_DATE_KEY: public_date})

    def _is_valid_file_metadata(self, metadata_dict: dict):
        """
        Check some file metadata for required headers.

        :param metadata_dict: dictionary of file metadata
        :return True if required headers are present, False if not
        """
        if any([k for k in self.required_headers if k not in metadata_dict]):
            return False
        else:
            return True

    def get_wcs_corners(self):
        """
        Returns a dict with a type and coordinates of the wcs coordinates if possible for this data type

        Take a fits dictionary and pick out the RA, DEC of each of the four corners.
        Then assemble a Polygon following the GeoJSON spec: http://geojson.org/geojson-spec.html#id4
        Note there are 5 positions. The last is the same as the first. We are defining lines,
        and you must close the polygon.

        If a RADIUS, RA, and DEC are set in the header, it will use that for a circular fov. Otherwise
        It will search for the standard header keys to build a polygon fov.
        """
        headers = self.header_data.get_headers()
        if self.header_data.headers_are_set([settings.RADIUS_KEY, settings.RA_KEY, settings.DEC_KEY]):
            ra = headers[settings.RA_KEY]
            dec = headers[settings.DEC_KEY]
            r = headers[settings.RADIUS_KEY]

            radius_in_degrees = Angle(r, units.arcsecond).deg
            ra_in_degrees = Angle(ra, units.hourangle).deg
            dec = Angle(dec, units.deg).deg

            c1 = (ra_in_degrees - radius_in_degrees, dec + radius_in_degrees)
            c2 = (ra_in_degrees + radius_in_degrees, dec + radius_in_degrees)
            c3 = (ra_in_degrees + radius_in_degrees, dec - radius_in_degrees)
            c4 = (ra_in_degrees - radius_in_degrees, dec - radius_in_degrees)

        elif (
                self.header_data.headers_are_set(['CD1_1', 'CD1_2', 'CD2_1', 'CD2_2', 'NAXIS1', 'NAXIS2'])
                and not self.header_data.headers_are_set(['NAXIS3'])
        ):
            # Find the RA and Dec coordinates of all 4 corners of the image
            try:
                w = wcs.WCS(headers)
                c1 = w.all_pix2world(1, 1, 1)
                c2 = w.all_pix2world(1, headers['NAXIS2'], 1)
                c3 = w.all_pix2world(headers['NAXIS1'], headers['NAXIS2'], 1)
                c4 = w.all_pix2world(headers['NAXIS1'], 1, 1)
            except ValueError:
                return None

        else:
            # This file doesn't have sufficient information to provide an area
            return None

        return {
            'type': 'Polygon',
            'coordinates': [
                [
                    [
                        float(c1[0]),
                        float(c1[1])
                    ],
                    [
                        float(c2[0]),
                        float(c2[1])
                    ],
                    [
                        float(c3[0]),
                        float(c3[1])
                    ],
                    [
                        float(c4[0]),
                        float(c4[1])
                    ],
                    [
                        float(c1[0]),
                        float(c1[1])
                    ]
                ]
            ]
        }

    def get_filestore_path(self):
        """Creates a path to use in the file store using the metadata of this file."""
        datatype = 'raw' if self.header_data.get_reduction_level() == 0 else 'processed'
        # Default directory is site/instrument/obsday/datatype/
        return '/'.join((self.header_data.get_site_id(), self.header_data.get_instrument_id(), self.header_data.get_observation_day(), datatype, self.open_file.basename)) + self.open_file.extension

    def get_filestore_content_type(self):
        return ''



from ocs_archive.storage.filestore import FileStore, FileStoreSpecificationError
from ocs_archive.storage.filesystemstore import FileSystemStore
from ocs_archive.storage.s3store import S3Store
from ocs_archive.settings import settings


FILESTORE_TYPES_TO_CLASS = {
    'dummy': FileStore,
    'local': FileSystemStore,
    's3': S3Store
}


class FileStoreFactory:
    @staticmethod
    def get_file_store_class(filestore_type: str = settings.FILESTORE_TYPE):
        if filestore_type not in FILESTORE_TYPES_TO_CLASS:
            raise FileStoreSpecificationError(f'Invalid FileStore type {filestore_type}')
        return FILESTORE_TYPES_TO_CLASS[filestore_type]



from ocs_archive.settings import settings


class HeaderData:
    def __init__(self, header_data: dict = None):
        """Creates HeaderData with input dictionary of headers."""
        if header_data is None:
            header_data = {}
        self.header_data = header_data

    def get_headers(self):
        return self.header_data

    def update_headers(self, new_headers: dict):
        self.header_data.update(new_headers)

    def remove_header(self, key: str):
        try:
            del self.header_data[key]
        except KeyError:
            pass

    def get_archive_frame_data(self):
        return {
            'reduction_level': self.get_reduction_level(),
            'observation_day': self.get_observation_day(),
            'observation_date': self.get_observation_date(),
            'proposal_id': self.get_proposal_id(),
            'instrument_id': self.get_instrument_id(),
            'target_name': self.get_target_name(),
            'site_id': self.get_site_id(),
            'telescope_id': self.get_telescope_id(),
            'exposure_time': self.get_exposure_time(),
            'primary_optical_element': self.get_primary_optical_element(),
            'public_date': self.get_public_date(),
            'configuration_type': self.get_configuration_type(),
            'observation_id': self.get_observation_id(),
            'request_id': self.get_request_id(),
            'related_frame_filenames': list(self.get_related_frames().values())
        }

    def headers_are_set(self, keys: list):
        """Check that the values for the provided headers are set."""
        headers = self.get_headers()
        empty_values = [None, '']
        values = [headers.get(key) for key in keys]
        return all([value not in empty_values for value in values])

    def get_related_frame_keys(self):
        return settings.RELATED_FRAME_KEYS

    def get_related_frames(self):
        headers = self.get_headers()
        keys = self.get_related_frame_keys()
        related_frames = {}
        for key in keys:
            if key in headers and headers[key]:
                related_frames[key] = headers[key]
        return related_frames

    def get_observation_day(self):
        return self.get_headers().get(settings.OBSERVATION_DAY_KEY)

    def get_observation_date(self):
        return self.get_headers().get(settings.OBSERVATION_DATE_KEY)

    def get_proposal_id(self):
        return self.get_headers().get(settings.PROPOSAL_ID_KEY, '')

    def get_configuration_type(self):
        return self.get_headers().get(settings.CONFIGURATION_TYPE_KEY, '')

    def get_exposure_time(self):
        return self.get_headers().get(settings.EXPOSURE_TIME_KEY)
    
    def get_public_date(self):
        return self.get_headers().get(settings.PUBLIC_DATE_KEY)

    def get_reduction_level(self):
        return self.get_headers().get(settings.REDUCTION_LEVEL_KEY, 0)

    def get_instrument_id(self):
        return self.get_headers().get(settings.INSTRUMENT_ID_KEY, '')

    def get_site_id(self):
        return self.get_headers().get(settings.SITE_ID_KEY, '')

    def get_primary_optical_element(self):
        return self.get_headers().get(settings.PRIMARY_OPTICAL_ELEMENT_KEY, '')

    def get_target_name(self):
        return self.get_headers().get(settings.TARGET_NAME_KEY, '')

    def get_telescope_id(self):
        return self.get_headers().get(settings.TELESCOPE_ID_KEY, '')

    def get_observation_id(self):
        return self.get_headers().get(settings.OBSERVATION_ID_KEY)

    def get_configuration_id(self):
        return self.get_headers().get(settings.CONFIGURATION_ID_KEY)

    def get_request_id(self):
        return self.get_headers().get(settings.REQUEST_ID_KEY)

    def get_requestgroup_id(self):
        return self.get_headers().get(settings.REQUESTGROUP_ID_KEY)



import tarfile
from contextlib import contextmanager

from ocs_archive.input.file import FileSpecificationException
from ocs_archive.input.fitsfile import FitsFile


class TarWithFitsFile(FitsFile):
    def _check_extension(self):
        if self.open_file.extension not in ['.tar', '.tar.gz']:
            raise FileSpecificationException(f'Tar with Fits files must have extension .tar or .tar.gz, not {self.file.extension}')

    @staticmethod
    def _get_meta_file_from_targz(tarfileobj):
        for member in tarfileobj.getmembers():
            if any(x in member.name for x in ['.fits', '.fits.fz']) and member.isfile():
                return tarfileobj.extractfile(member)
        raise FileSpecificationException('Tar file missing meta fits!')

    @contextmanager
    def get_fits(self):
        """
        Return the fits file associated with this file.

        Generally this just the fileobj itself, but certain spectral data must have their
        fits files extracted. Use this as a context manager.
        """
        fits_file, tar_file = None, None
        try:
            tar_file = tarfile.open(fileobj=self.open_file.get_from_start(), mode='r')
            fits_file = self._get_meta_file_from_targz(tar_file)
            yield fits_file

        finally:
            if tar_file and fits_file:
                # The fits file object came from the tar file extraction, and must be closed
                fits_file.close()
            if tar_file:
                tar_file.close()

    def get_filestore_content_type(self):
        return 'application/x-tar'



from io import BytesIO
from contextlib import contextmanager

from ocs_archive.input.file import DataFile


class FileStoreConnectionError(Exception):
    """
    This exception should generally be caught and retried in the calling class,
    since it may just be a transient connection issue.
    """
    pass


class FileStoreSpecificationError(Exception):
    """This exception is for an invalid configured filestore, and should not be retried."""
    pass


class FileStore:
    def store_file(self, data_file:DataFile):
        # This should store the DataFile wherever is relevant for the implementing class
        # It should return the version_set information, which is a dictionary containing
        # a unique version key for the stored file, the md5 of the stored file, and the
        # files extension
        md5 = data_file.open_file.get_md5()
        return {
            'key': md5,
            'md5': md5,
            'extension': data_file.open_file.extension
        }

    def delete_file(self, path: str, version_id: str):
        pass

    def get_url(self, path: str, version_id: str, expiration: float):
        return ''

    @contextmanager
    def get_fileobj(self, path: str):
        try:
            fileobj = BytesIO()
            yield fileobj
        finally:
            fileobj.close()

    def get_file_size(self, path: str):
        return 0



from astropy.io import fits
import io
from contextlib import contextmanager

from ocs_archive.input.file import File, DataFile, FileSpecificationException
from ocs_archive.input.headerdata import HeaderData
from ocs_archive.settings import settings


class FitsFile(DataFile):
    # These are header keys that must contain integer values or None for archive validation
    INTEGER_TYPES = [
        settings.REQUEST_ID_KEY, settings.REQUESTGROUP_ID_KEY,
        settings.OBSERVATION_ID_KEY, settings.CONFIGURATION_ID_KEY
    ]

    def __init__(self, open_file: File, file_metadata: dict = None, blacklist_headers: tuple = settings.HEADER_BLACKLIST, required_headers: tuple = settings.REQUIRED_HEADERS):
        """Loads in fits file headers, then does some automatic cleanup and normalization of values."""
        super().__init__(open_file, file_metadata, blacklist_headers, required_headers)
        self._check_extension()
        self._remove_blacklist_headers(blacklist_headers)
        self._normalize_null_values()
        self._normalize_related_frames()

    def _remove_blacklist_headers(self, blacklist_headers: tuple):
        if '' not in blacklist_headers:
            # Always remove the empty string header since it causes issues
            self.header_data.remove_header('')
        for header in self.blacklist_headers:
            self.header_data.remove_header(header)

    def _normalize_null_values(self):
        header_updates = {}
        #  Sometimes keywords use N/A to mean null
        for k, v in self.header_data.get_headers().items():
            if v in settings.NULL_HEADER_VALUES:
                if k in self.INTEGER_TYPES:
                    header_updates[k] = None
                else:
                    header_updates[k] = ''
            # Catch None and NONE values for Integer type fields so they pass archive validation
            elif k in self.INTEGER_TYPES and ('NONE' in str(v).upper()):
                header_updates[k] = None
        self.header_data.update_headers(header_updates)

    def _normalize_related_frames(self):
        related_frame_keys = self.header_data.get_related_frame_keys()
        headers = self.header_data.get_headers()
        for key in related_frame_keys:
            filename = headers.get(key)
            if filename and filename != 'N/A':
                basename, extension = File.get_basename_and_extension(filename)
                if extension:
                    # Remove any extensions for the related frames, since the archive expects that
                    self.header_data.update_headers({key: basename})

    @contextmanager
    def get_fits(self):
        yield self.open_file.get_from_start()

    def _check_extension(self):
        if self.open_file.extension not in ['.fits', '.fits.fz']:
            raise FileSpecificationException(f'Fits files must have extension .fits or .fits.fz, not {self.open_file.extension}')

    def _is_valid_fits(self):
        try:
            with self.get_fits() as fits_file:
                with fits.open(io.BytesIO(fits_file.read()), mode='readonly') as hdu_list:
                    hdu_list.verify()
                return True
        except Exception:
            return False

    def _create_header_data(self, file_metadata: dict):
        if self._is_valid_fits():
            # Loop through each HDU and use the first header that passes validation as the dict representation
            with self.get_fits() as fits_file:
                with fits.open(io.BytesIO(fits_file.read()), mode='readonly') as hdulist:
                    for hdu in hdulist:
                        fits_dict = dict(hdu.header)
                        if file_metadata:
                            fits_dict.update(file_metadata)
                        if self._is_valid_file_metadata(fits_dict):
                            self.header_data = HeaderData(fits_dict)
                            return
                        else:
                            continue
                raise FileSpecificationException(
                    'Could not find required keywords in headers!')  # Missing one or more required headers
        else:
            # If there are no valid fits headers in the file, fall back on passed in metadata
            super()._create_header_data(file_metadata)

    def get_filestore_content_type(self):
        return 'image/fits'



from datetime import datetime, timedelta
import logging
from io import BytesIO
from functools import lru_cache
from contextlib import contextmanager

from dateutil.parser import parse
import boto3

from ocs_archive.input.file import DataFile
from ocs_archive.storage.filestore import FileStore, FileStoreConnectionError
from ocs_archive.settings import settings

logger = logging.getLogger('ocs_ingester')


def strip_quotes_from_etag(etag):
    """Amazon returns the md5 sum of the uploaded file in the 'ETag' header wrapped in quotes."""
    if etag.startswith('"') and etag.endswith('"'):
        return etag[1:-1]


class S3Store(FileStore):
    def __init__(self, bucket: str = settings.BUCKET):
        """Create an S3 file storage manager using the bucket specified."""
        super().__init__()
        self.bucket = bucket

    @classmethod
    @lru_cache(maxsize=1)
    def get_s3_client(cls):
        config = boto3.session.Config(signature_version=settings.S3_SIGNATURE_VERSION, s3={'addressing_style': 'virtual'})
        return boto3.client('s3', endpoint_url=settings.S3_ENDPOINT_URL, config=config)

    def get_storage_class(self, observation_date):
        # if the observation was more than X days ago, this is someone
        # uploading older data, and it can skip straight to STANDARD_IA
        if observation_date < (datetime.utcnow() - timedelta(days=settings.S3_DAYS_TO_IA_STORAGE)):
            return 'STANDARD_IA'

        # everything else goes into the STANDARD storage class, and will
        # be switched to STANDARD_IA by S3 Lifecycle Rules
        return 'STANDARD'

    def store_file(self, data_file: DataFile):
        storage_class = self.get_storage_class(parse(data_file.get_header_data().get_observation_date()))
        # start_time = datetime.utcnow()
        s3 = boto3.resource('s3', endpoint_url=settings.S3_ENDPOINT_URL)
        key = data_file.get_filestore_path()
        content_disposition = 'attachment; filename={0}{1}'.format(data_file.open_file.basename, data_file.open_file.extension)
        content_type = data_file.get_filestore_content_type()
        try:
            response = s3.Object(self.bucket, key).put(
                Body=data_file.open_file.get_from_start(),
                ContentDisposition=content_disposition,
                ContentType=content_type,
                StorageClass=storage_class,
            )
        except Exception as exc:
            raise FileStoreConnectionError(exc)
        s3_md5 = strip_quotes_from_etag(response['ETag'])
        key = response['VersionId']
        logger.info('Ingester uploaded file to s3', extra={
            'tags': {
                'filename': '{}{}'.format(data_file.open_file.basename, data_file.open_file.extension),
                'key': key,
                'storage_class': storage_class,
            }
        })

        return {'key': key, 'md5': s3_md5, 'extension': data_file.open_file.extension}

    def delete_file(self, path: str, version_id: str):
        """
        Delete a file from s3.

        :param path: s3 path for file
        """
        client = S3Store.get_s3_client()
        client.delete_object(
            Bucket=self.bucket,
            Key=path,
            VersionId=version_id
        )

    def get_url(self, path: str, version_id: str, expiration: float):
        """
        Get a downloadable url for a file from s3.

        :param path: s3 path for file
        """
        client = S3Store.get_s3_client()
        return client.generate_presigned_url(
            'get_object',
            ExpiresIn=expiration,
            Params={
                'Bucket': self.bucket,
                'Key': path,
                'VersionId': version_id
            }
        )

    @contextmanager
    def get_fileobj(self, path: str):
        """
        Get a file from s3.

        DataFile can contain an EmptyFile within it, but the headers must be correct
        To figure out the file storage path.

        :param data_file: DataFile with filled in headers pointing to storage path
        :return: File-like object
        """
        client = S3Store.get_s3_client()
        fileobj = BytesIO()
        client.download_fileobj(Bucket=self.bucket,
                                Key=path,
                                Fileobj=fileobj)
        fileobj.seek(0)
        try:
            yield fileobj
        finally:
            fileobj.close()

    def get_file_size(self, path: str):
        """
        Get the size of a file in s3.

        :param path: s3 path for file
        :return: file size in bytes
        """
        client = S3Store.get_s3_client()
        return client.head_object(Bucket=self.bucket, Key=path)['ContentLength']



from importlib import import_module

from ocs_archive.input.file import DataFile, FileSpecificationException
from ocs_archive.settings import settings

EXTENSION_TO_FILE_CLASS = {
    '.fits.fz': 'ocs_archive.input.lcofitsfile.LcoFitsFile',
    '.fits': 'ocs_archive.input.fitsfile.FitsFile',
    '.tar.gz': 'ocs_archive.input.tarwithfitsfile.TarWithFitsFile',
    '.pdf': 'ocs_archive.input.file.DataFile'
}


class FileFactory:
    """
    This factory returns the proper datafile subclass based on the file extension.
    It has a default mapping defined above, but that can be extended and overridden
    via the FILETYPE_MAPPING_OVERRIDES environment variable.
    """
    @staticmethod
    def get_datafile_class_for_extension(extension: str):
        extensions_dict = EXTENSION_TO_FILE_CLASS.copy()
        if isinstance(settings.FILETYPE_MAPPING_OVERRIDES, dict):
            extensions_dict.update(settings.FILETYPE_MAPPING_OVERRIDES)
        if extension not in extensions_dict:
            raise FileSpecificationException(f'file extension {extension} is not a currently supported file type')

        module_path, _, class_name = extensions_dict[extension].rpartition('.')
        try:
            loaded_module = import_module(module_path)
        except ModuleNotFoundError:
            raise FileSpecificationException(f'module {module_path} was not found. Please ensure the class path is correct')

        try:
            file_class = getattr(loaded_module, class_name)
        except AttributeError:
            raise FileSpecificationException(f'class {class_name} does not exist within module {module_path}. Please ensure the class path is correct')

        if not issubclass(file_class, DataFile):
            raise FileSpecificationException(f'class {class_name} must be a subclass of the ocs_archive.input.file.DataFile class')

        return file_class



from ocs_archive.input.fitsfile import FitsFile
from ocs_archive.input.file import File, FileSpecificationException
from ocs_archive.settings import settings


LETTER_TO_OBSTYPE = {
    'b': 'BIAS',
    'd': 'DARK',
    'f': 'SKYFLAT',
    'g': 'GUIDE',
    's': 'STANDARD',
    'w': 'LAMPFLAT',
    'x': 'EXPERIMENTAL'
}

class LcoFitsFile(FitsFile):
    def __init__(self, open_file: File, file_metadata: dict = None, blacklist_headers: tuple = settings.HEADER_BLACKLIST, required_headers: tuple = settings.REQUIRED_HEADERS):
        """LCO specific FitsFile format, provides some extra repair for broken headers from the filename."""
        super().__init__(open_file, file_metadata, blacklist_headers, required_headers)
        self._check_catalog()
        self._repair_configuration_type()
        self._repair_reduction_level()

    def _check_catalog(self):
        if '_cat' in self.open_file.basename:
            header_updates = {}
            headers = self.header_data.get_headers()
            if not headers.get(settings.CATALOG_TARGET_FRAME_KEY):
                # Check if the catalog file contains it's target frame, if not deduce it
                l1idcat = self.open_file.basename.replace('_cat', '')
                header_updates[settings.CATALOG_TARGET_FRAME_KEY] = l1idcat
            # set configuration type to CATALOG even though it's might be set to EXPOSE by the pipeline
            # TODO: pipeline should just create catalog files with correct keys
            header_updates[settings.CONFIGURATION_TYPE_KEY] = 'CATALOG'
            self.header_data.update_headers(header_updates)

    def _repair_configuration_type(self):
        conftype = self.header_data.get_configuration_type()
        if not conftype or conftype.strip() == 'UNKNOWN':
            try:
                headers = self.header_data.get_headers()
                name_parts = self.open_file.basename.split('-')
                conftype_letter = name_parts[4][0]
                is_nres = 'igl' in headers.get('ENCID', '') or (self.header_data.get_telescope_id() and 'igl' in self.header_data.get_telescope_id())
                if 'trace' == name_parts[0]:
                    conftype = 'TRACE'
                elif 'arc' == name_parts[0]:
                    conftype = 'ARC'
                elif 'bias' == name_parts[3]:
                    conftype = 'BIAS'
                elif 'bpm' == name_parts[3]:
                    conftype = 'BPM'
                elif conftype_letter == 'e':
                    if is_nres:
                        conftype = 'TARGET'
                    elif 'en' in name_parts[1]:
                        conftype = 'SPECTRUM'
                    else:
                        conftype = 'EXPOSE'
                elif conftype_letter == 'a':
                    if is_nres:
                        conftype = 'DOUBLE'
                    else:
                        conftype = 'ARC'
                elif conftype_letter in LETTER_TO_OBSTYPE:
                    conftype = LETTER_TO_OBSTYPE[conftype_letter]
                else:
                    # Failed to infer an configuration type for this filename
                    raise Exception()
                # Set the configuration type into the header
                self.header_data.update_headers({settings.CONFIGURATION_TYPE_KEY: conftype})
            except Exception:
                raise FileSpecificationException(f'{settings.CONFIGURATION_TYPE_KEY} is UNKNOWN and could not be inferred. Please manually correct it')

    def _repair_reduction_level(self):
        reduction_level = self.header_data.get_reduction_level()
        if not reduction_level:
            # TODO: Pipeline should write this value instead of
            # being inferred from the filename
            MAX_REDUCTION = 90
            MIN_REDUCTION = 0

            # remove the _cat extension from catalog files
            basename = self.open_file.basename.replace('_cat', '')

            if self.open_file.extension == '.tar.gz':
                self.header_data.update_headers({settings.REDUCTION_LEVEL_KEY: MAX_REDUCTION})
            else:
                try:
                    # lsc1m005-kb78-20151007-0214-x00
                    # extract reduction level at the position of 00
                    self.header_data.update_headers({settings.REDUCTION_LEVEL_KEY: int(basename[-2:])})
                except ValueError:
                    # Some filenames don't have this extension - return a sensible default
                    self.header_data.update_headers({settings.REDUCTION_LEVEL_KEY: MIN_REDUCTION})

    def _is_bpm_file(self):
        """Check if the file is a bad pixel mask using several LCO specific rules"""
        headers = self.header_data.get_headers()
        if self.header_data.get_configuration_type() == 'BPM' or headers.get('EXTNAME') == 'BPM':
            return True
        test_filename = self.open_file.basename.replace('_', '-')
        if test_filename.startswith('bpm-') or '-bpm-' in test_filename or test_filename.endswith('-bpm'):
            return True
        return False

    def get_filestore_path(self):
        if self._is_bpm_file():
            return '/'.join((self.header_data.get_site_id(), self.header_data.get_instrument_id(), 'bpm', self.open_file.basename)) + self.open_file.extension
        return super().get_filestore_path()


import os
import ast


def get_tuple_from_environment(variable_name, default):
    return tuple(os.getenv(variable_name, default).strip(',').replace(' ', '').split(','))


# AWS Credentials and defaults
BUCKET = os.getenv('BUCKET', 'testbucket')
S3_SIGNATURE_VERSION = os.getenv('S3_SIGNATURE_VERSION', 's3v4')
S3_ENDPOINT_URL = os.getenv('S3_ENDPOINT_URL', 'https://s3.us-west-2.amazonaws.com')
S3_DAYS_TO_IA_STORAGE = int(os.getenv('S3_DAYS_TO_IA_STORAGE', 60))

# Used to specify which type of File Storage to use
FILESTORE_TYPE = os.getenv('FILESTORE_TYPE', 'dummy')

# Used for local File system storage backend as the root storage directory
FILESYSTEM_STORAGE_ROOT_DIR = os.getenv('FILESYSTEM_STORAGE_ROOT_DIR', '')
FILESYSTEM_STORAGE_BASE_URL = os.getenv('FILESYSTEM_STORAGE_BASE_URL', 'http://0.0.0.0/')

# Used to specify the location and authentication details for the Observation Portal
OBSERVATION_PORTAL_BASE_URL = os.getenv('OBSERVATION_PORTAL_BASE_URL')
OBSERVATION_PORTAL_API_TOKEN = os.getenv('OBSERVATION_PORTAL_API_TOKEN')

# Used to specify the proposal tags that denote whether the data from a proposal should be public or private
PRIVATE_PROPOSAL_TAGS = get_tuple_from_environment('PRIVATE_PROPOSAL_TAGS', 'private,internal')
PUBLIC_PROPOSAL_TAGS = get_tuple_from_environment('PUBLIC_PROPOSAL_TAGS', 'public')

# Used to override and update mapping of file extensions to DataFile subclass class dotpath
# The expected format is a string literal representation of a dictionary, mapping extensions
# to Datafile subclass absolute dotpath
FILETYPE_MAPPING_OVERRIDES = ast.literal_eval(os.getenv('FILETYPE_MAPPING_OVERRIDES', "{}"))

# Files we wish to ignore
IGNORED_CHARS = get_tuple_from_environment('IGNORED_CHARS', '-l00,tstnrs')

# Fits headers we don't want to ingest
HEADER_BLACKLIST = get_tuple_from_environment('HEADER_BLACKLIST', 'HISTORY,COMMENT')

# Fits headers that must be present
REQUIRED_HEADERS = get_tuple_from_environment('REQUIRED_HEADERS', 'PROPID,DATE-OBS,INSTRUME,SITEID,TELID,OBSTYPE,BLKUID')

# Possible Null values in headers. These will be normalized to none / empty
NULL_HEADER_VALUES = get_tuple_from_environment('NULL_HEADER_VALUES', 'N/A,UNSPECIFIED,UNKNOWN')

# Calibration observation types (OBSTYPE)
CALIBRATION_TYPES = get_tuple_from_environment('CALIBRATION_TYPES', 'BIAS,DARK,SKYFLAT,EXPERIMENTAL')

# Proposals including these strings will be considered public data
PUBLIC_PROPOSALS = get_tuple_from_environment('PUBLIC_PROPOSALS', 'EPO,calib,standard,pointing')

# Days until a private proposals data becomes public, measured from observation date
DAYS_UNTIL_PUBLIC = int(os.getenv('DAYS_UNTIL_PUBLIC', 365))

# Proposals including these strings will be considered private data (L1PUBDATE far out)
PRIVATE_PROPOSALS = get_tuple_from_environment('PRIVATE_PROPOSALS', 'LCOEngineering')

# File types which are private (L1PUBDATE far out)
PRIVATE_FILE_TYPES = get_tuple_from_environment('PRIVATE_FILE_TYPES', '-t00,-x00')

# Data File related settings (default keys, related frames, etc.)
OBSERVATION_DATE_KEY = os.getenv('OBSERVATION_DATE_KEY', 'DATE-OBS')
OBSERVATION_DAY_KEY = os.getenv('OBSERVATION_DAY_KEY', 'DAY-OBS')
OBSERVATION_END_TIME_KEY = os.getenv('OBSERVATION_END_TIME_KEY', 'UTSTOP')
REDUCTION_LEVEL_KEY = os.getenv('REDUCTION_LEVEL_KEY', 'RLEVEL')
EXPOSURE_TIME_KEY = os.getenv('EXPOSURE_TIME_KEY', 'EXPTIME')
INSTRUMENT_ID_KEY = os.getenv('INSTRUMENT_ID_KEY', 'INSTRUME')
SITE_ID_KEY = os.getenv('SITE_ID_KEY', 'SITEID')
TELESCOPE_ID_KEY = os.getenv('TELESCOPE_ID_KEY', 'TELID')
OBSERVATION_ID_KEY = os.getenv('OBSERVATION_ID_KEY', 'BLKUID')
CONFIGURATION_ID_KEY = os.getenv('CONFIGURATION_ID_KEY', 'MOLUID')
PRIMARY_OPTICAL_ELEMENT_KEY = os.getenv('PRIMARY_OPTICAL_ELEMENT_KEY', 'FILTER')
TARGET_NAME_KEY = os.getenv('TARGET_NAME_KEY', 'OBJECT')
REQUEST_ID_KEY = os.getenv('REQUEST_ID_KEY', 'REQNUM')
REQUESTGROUP_ID_KEY = os.getenv('REQUESTGROUP_ID_KEY', 'TRACKNUM')
CONFIGURATION_TYPE_KEY = os.getenv('CONFIGURATION_TYPE_KEY', 'OBSTYPE')
PROPOSAL_ID_KEY = os.getenv('PROPOSAL_ID_KEY', 'PROPID')
CATALOG_TARGET_FRAME_KEY = os.getenv('CATALOG_TARGET_FRAME_KEY', 'L1IDCAT')
PUBLIC_DATE_KEY = os.getenv('PUBLIC_DATE_KEY', 'L1PUBDAT')
RELATED_FRAME_KEYS = get_tuple_from_environment(
    'RELATED_FRAME_KEYS',
    'L1IDBIAS,L1IDDARK,L1IDFLAT,L1IDSHUT,L1IDMASK,L1IDFRNG,L1IDCAT,L1IDARC,L1ID1D,L1ID2D,L1IDSUM,TARFILE,ORIGNAME,ARCFILE,FLATFILE,GUIDETAR'
)
# Either Radius, Ra, Dec or NAXIS1/2 and CD1/2_1/2 need to be set to support
# automatic extraction of wcs polygon
RADIUS_KEY = os.getenv('RADIUS_KEY', 'RADIUS')
RA_KEY = os.getenv('RA_KEY', 'RA')
DEC_KEY = os.getenv('DEC_KEY', 'DEC')
